"""\
Example.

    %(prog)s production.ini

"""

import argparse
import atexit
import datetime
import elasticsearch.exceptions
import json
import logging
import os
import psycopg2
import signal
import sqlalchemy.exc
import structlog
import sys
import threading
import time
import webtest

from pyramid import paster
from dcicutils.log_utils import set_logging
from .interfaces import ELASTIC_SEARCH, INDEXER_QUEUE


log = structlog.getLogger(__name__)

EPILOG = __doc__
DEFAULT_INTERVAL = 3  # 3 second default

# We need this because of MVCC visibility.
# See slide 9 at http://momjian.us/main/writings/pgsql/mvcc.pdf
# https://devcenter.heroku.com/articles/postgresql-concurrency


def run(testapp, interval=DEFAULT_INTERVAL, dry_run=False, path='/index', update_status=None):
    log.info('___INDEXER LISTENER STARTING___')
    listening = False
    timestamp = datetime.datetime.now().isoformat()
    update_status(
        listening=listening,
        status='indexing',
        timestamp=timestamp
    )

    # Make sure elasticsearch is up before trying to index.
    es = testapp.app.registry[ELASTIC_SEARCH]
    es.info()

    queue = testapp.app.registry[INDEXER_QUEUE]

    # main listening loop
    while True:
        # if not messages to index, skip the /index call. Counts are approximate
        queue_counts = queue.number_of_messages()
        if (not queue_counts['primary_waiting'] and not queue_counts['secondary_waiting']):
            time.sleep(interval)
            continue

        try:
            res = testapp.post_json(path, {
                'record': True,
                'dry_run': dry_run
            })
        except Exception as e:
            timestamp = datetime.datetime.now().isoformat()
            log.exception('index failed')
            update_status(error={
                'error': repr(e),
                'timestamp': timestamp,
            })
        else:
            timestamp = datetime.datetime.now().isoformat()
            result = res.json
            result['stats'] = res.headers.get('X-Stats', {})
            result['timestamp'] = timestamp
            update_status(last_result=result)
            if result.get('indexing_status') == 'finished':
                update_status(result=result)
                if result.get('errors'):
                    log.error('___INDEX LISTENER RESULT:___\n%s\n' % result)
                else:
                    log.debug('___INDEX LISTENER RESULT:___\n%s\n' % result)
        time.sleep(interval)


class ErrorHandlingThread(threading.Thread):

    def run(self):
        # interval = self._kwargs.get('interval', DEFAULT_INTERVAL)
        interval = 60  # DB polling can and should be slower
        update_status = self._kwargs['update_status']
        while True:
            try:
                self._target(*self._args, **self._kwargs)
            except (psycopg2.OperationalError, sqlalchemy.exc.OperationalError, elasticsearch.exceptions.ConnectionError) as e:
                # Handle database restart
                log.warning('Database not there, maybe starting up: %r', e)
                timestamp = datetime.datetime.now().isoformat()
                update_status(
                    timestamp=timestamp,
                    status='sleeping',
                    error={'error': repr(e), 'timestamp': timestamp},
                )
                log.debug('sleeping')
                time.sleep(interval)
                continue
            except Exception:
                # Unfortunately mod_wsgi does not restart immediately
                log.exception('Exception in listener, restarting process at next request.')
                os.kill(os.getpid(), signal.SIGINT)
            break


def composite(loader, global_conf, **settings):
    listener = None

    # Register before testapp creation.
    @atexit.register
    def join_listener():
        if listener:
            log.debug('joining listening thread')
            listener.join()

    path = settings.get('path', '/index')

    # Composite app is used so we can load the main app
    app_name = settings.get('app', None)
    app = loader.get_app(app_name, global_conf=global_conf)
    username = settings.get('username', 'IMPORT')
    environ = {
        'HTTP_ACCEPT': 'application/json',
        'REMOTE_USER': username,
    }
    testapp = webtest.TestApp(app, environ)


    timestamp = datetime.datetime.now().isoformat()
    status_holder = {
        'status': {
            'status': 'starting listener',
            'started': timestamp,
            'errors': [],
            'results': [],
        },
    }

    def update_status(error=None, result=None, indexed=None, **kw):
        # Setting a value in a dictionary is atomic
        status = status_holder['status'].copy()
        status.update(**kw)
        if error is not None:
            status['errors'] = [error] + status['errors'][:2]
        if result is not None:
            status['results'] = [result] + status['results'][:9]
        status_holder['status'] = status

    kwargs = {
        'testapp': testapp,
        'update_status': update_status,
        'path': path,
    }
    if 'interval' in settings:
        kwargs['interval'] = float(settings['interval'])

    # daemon thread that actually executes `run` method to call /index
    listener = ErrorHandlingThread(target=run, name='listener', kwargs=kwargs)
    listener.daemon = True
    log.debug('starting listener')
    listener.start()

    # Register before testapp creation.
    @atexit.register
    def shutdown_listener():
        log.debug('shutting down listening thread')

    def status_app(environ, start_response):
        status = '200 OK'
        response_headers = [('Content-type', 'application/json')]
        start_response(status, response_headers)
        return [json.dumps(status_holder['status'])]

    return status_app


def internal_app(configfile, app_name=None, username=None):
    app = paster.get_app(configfile, app_name)
    if not username:
        username = 'IMPORT'
    environ = {
        'HTTP_ACCEPT': 'application/json',
        'REMOTE_USER': username,
    }
    return webtest.TestApp(app, environ)


def main():
    parser = argparse.ArgumentParser(
        description="Listen for changes from postgres and index in elasticsearch",
        epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('--app-name', help="Pyramid app name in configfile")
    parser.add_argument(
        '--username', '-u', default='INDEXER', help="Import username")
    parser.add_argument(
        '--dry-run', action='store_true', help="Don't post to ES, just print")
    parser.add_argument(
        '-v', '--verbose', action='store_true', help="Print debug level logging")
    parser.add_argument(
        '--poll-interval', type=int, default=DEFAULT_INTERVAL,
        help="Poll interval between notifications")
    parser.add_argument(
        '--path', default='/index',
        help="Path of indexing view")
    parser.add_argument('config_uri', help="path to configfile")
    args = parser.parse_args()

    # logging.basicConfig()
    testapp = internal_app(args.config_uri, args.app_name, args.username)


    # Loading app will have configured from config file. Reconfigure here:
    level = logging.INFO
    if args.verbose or args.dry_run:
        level = logging.DEBUG

    # Loading app will have configured from config file. Reconfigure here:
    # Use `es_server=app.registry.settings.get('elasticsearch.server')` when ES logging is working
    set_logging(in_prod=testapp.app.registry.settings.get('production'), level=logging.INFO)
    return run(testapp, args.poll_interval, args.dry_run, args.path)


if __name__ == '__main__':
    main()



"""\
Also run this when the links or keys are changed in the schema.

Example:

    %(prog)s production.ini --app-name app

"""

import argparse
import itertools
import transaction as transaction_management
import webtest

from copy import deepcopy
# multiprocessing.get_context is Python 3 only.
from multiprocessing import get_context
from multiprocessing.pool import Pool
from pyramid import paster
from pyramid.traversal import find_resource
from pyramid.view import view_config
from structlog import getLogger

from .interfaces import CONNECTION, STORAGE, UPGRADER

from .schema_utils import validate
from .util import debug_log


logger = getLogger(__name__)


EPILOG = __doc__


def includeme(config):
    config.add_route('batch_upgrade', '/batch_upgrade')
    config.scan(__name__)


def batched(iterable, n=1):
    l = len(iterable)
    for ndx in range(0, l, n):
        yield iterable[ndx:min(ndx+n, l)]


def internal_app(configfile, app_name=None, username=None):
    app = paster.get_app(configfile, app_name)
    if not username:
        username = 'UPGRADE'
    environ = {
        'HTTP_ACCEPT': 'application/json',
        'REMOTE_USER': username,
    }
    return webtest.TestApp(app, environ)


# Running in subprocess
testapp = None


def initializer(*args, **kw):
    global testapp
    testapp = internal_app(*args, **kw)


def worker(batch):
    res = testapp.post_json('/batch_upgrade', {'batch': batch})
    return res.json


def update_item(storage, context):
    target_version = context.type_info.schema_version
    current_version = context.properties.get('schema_version', '1')
    update = False
    errors = []
    properties = context.properties
    if target_version is None or current_version == target_version:
        unique_keys = context.unique_keys(properties)
        links = context.links(properties)
        keys_add, keys_remove = storage._update_keys(context.model, unique_keys)
        if keys_add or keys_remove:
            update = True
        rels_add, rels_remove = storage._update_rels(context.model, links)
        if rels_add or rels_remove:
            update = True
    else:
        properties = deepcopy(properties)
        upgrader = context.registry[UPGRADER]
        properties = upgrader.upgrade(
            context.type_info.name, properties, current_version, target_version,
            context=context, registry=context.registry)
        if 'schema_version' in properties:
            del properties['schema_version']
        schema = context.type_info.schema
        properties['uuid'] = str(context.uuid)
        validated, errors = validate(schema, properties, properties)
        # Do not send modification events to skip indexing
        context.update(validated)
        update = True
    return update, errors


@view_config(route_name='batch_upgrade', request_method='POST', permission='import_items')
@debug_log
def batch_upgrade(request):
    request.datastore = 'database'
    transaction_management.get().setExtendedInfo('upgrade', True)
    batch = request.json['batch']
    root = request.root
    storage = request.registry[STORAGE].write
    session = storage.DBSession()
    results = []
    for uuid in batch:
        item_type = None
        update = False
        error = False
        sp = session.begin_nested()
        try:
            item = find_resource(root, uuid)
            item_type = item.type_info.item_type
            update, errors = update_item(storage, item)
        except Exception:
            logger.exception('Error updating: /%s/%s', item_type, uuid)
            sp.rollback()
            error = True
        else:
            if errors:
                errortext = [
                    '%s: %s' % ('/'.join(error.path) or '<root>', error.message)
                    for error in errors]
                logger.error(
                    'Validation failure: /%s/%s\n%s', item_type, uuid, '\n'.join(errortext))
                sp.rollback()
                error = True
            else:
                sp.commit()
        results.append((item_type, uuid, update, error))
    return {'results': results}


def run(config_uri, app_name=None, username=None, types=(), batch_size=500, processes=None):
    # Loading app will have configured from config file. Reconfigure here:
    logging.getLogger('snovault').setLevel(logging.DEBUG)

    testapp = internal_app(config_uri, app_name, username)
    connection = testapp.app.registry[CONNECTION]
    uuids = [str(uuid) for uuid in connection.__iter__(*types)]
    transaction_management.abort()
    logger.info('Total items: %d' % len(uuids))

    pool = Pool(
        processes=processes,
        initializer=initializer,
        initargs=(config_uri, app_name, username),
        context=get_context('forkserver'),
    )

    all_results = []
    try:
        for result in pool.imap_unordered(worker, batched(uuids, batch_size), chunksize=1):
            results = result['results']
            errors = sum(error for item_type, path, update, error in results)
            updated = sum(update for item_type, path, update, error in results)
            logger.info('Batch: Updated %d of %d (errors %d)' %
                        (updated, len(results), errors))
            all_results.extend(results)
    finally:
        pool.terminate()
        pool.join()

    def result_item_type(result):
        # Ensure we always return a string
        return result[0] or ''

    for item_type, results in itertools.groupby(
            sorted(all_results, key=result_item_type), key=result_item_type):
        results = list(results)
        errors = sum(error for item_type, path, update, error in results)
        updated = sum(update for item_type, path, update, error in results)
        logger.info('Collection %s: Updated %d of %d (errors %d)' %
                    (item_type, updated, len(results), errors))


def main():
    parser = argparse.ArgumentParser(
        description="Batch upgrade content items.", epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('config_uri', help="path to configfile")
    parser.add_argument('--app-name', help="Pyramid app name in configfile")
    parser.add_argument('--item-type', dest='types', action='append', default=[])
    parser.add_argument('--batch-size', type=int, default=1000)
    parser.add_argument('--processes', type=int)
    parser.add_argument('--username')
    args = parser.parse_args()
    logging.basicConfig()
    run(**vars(args))


if __name__ == '__main__':
    main()



import pytest
import mock
import copy
from contextlib import contextmanager
from ..elasticsearch.indexer_utils import filter_invalidation_scope, determine_child_types


# Mocked uuids
UUID1 = 'UUID1'
UUID2 = 'UUID2'

# Mocked item types
ITEM_A = 'Item_A'
ITEM_B = 'Item_B'
ITEM_C = 'Item_C'
ITEM_D = 'Item_D'
ITEM_E = 'Item_E'
ITEM_F = 'Item_F'


@contextmanager
def type_properties_mock(schema):
    with mock.patch('snovault.elasticsearch.indexer_utils.extract_type_properties',
                    return_value=schema):
        yield


@contextmanager
def type_embedded_list_mock(embedded_list):
    with mock.patch('snovault.elasticsearch.indexer_utils.extract_type_embedded_list',
                    return_value=embedded_list):
        yield


@contextmanager
def type_default_diff_mock(use_default_diff):
    if use_default_diff:
        yield
    else:
        with mock.patch('snovault.elasticsearch.indexer_utils.extract_type_default_diff',
                        return_value=[]):
            yield


@contextmanager
def type_base_types_mock(base_types):
    if base_types is not None:
        with mock.patch('snovault.elasticsearch.indexer_utils.extract_base_types',
                        return_value=base_types):
            yield
    else:
        yield


@contextmanager
def type_child_types_mock(child_types):
    if child_types is not None:
        with mock.patch('snovault.elasticsearch.indexer_utils.determine_child_types',
                        return_value=child_types):
            yield
    else:
        yield


@contextmanager
def invalidation_scope_mocks(*, schema, embedded_list, base_types=None, child_types=None, use_default_diff=False):
    """ Quick wrapper for a common operation in this testing (mocking the appropriate things to test this
        without a data model) """
    with type_properties_mock(schema):
        with type_embedded_list_mock(embedded_list):
            with type_default_diff_mock(use_default_diff):
                with type_base_types_mock(base_types):
                    with type_child_types_mock(child_types):
                        yield


@pytest.fixture
def test_link_source_schema():
    """ Intended schema structure. Meant to test multiple different modification scenarios
            Item_A (uuid1):
                * link_one: linkTo C
                * link_two: linkTo D
                * link_three: linkTo E
                * link_four: array of linkTo F
    """
    return {
        'link_one': {
            'type': 'string',
            'linkTo': ITEM_C
        },
        'link_two': {
            'type': 'string',
            'linkTo': ITEM_D
        },
        'link_three': {
            'type': 'object',
            'linkTo': ITEM_E
        },
        'link_four': {
            'type': 'array',
            'items': {
                'type': 'string',
                'linkTo': ITEM_F
            }
        }
    }


@pytest.fixture
def test_parent_type_schema():
    """ Intended schema structure to test schemas that linkTo base types
            Item_B (uuid2):
                * link_one: linkTo C where C is a parent class of D
                * link_two: linkTo E where E is a parent class of F

    """
    return {
        'link_one': {
            'type': 'string',
            'linkTo': ITEM_C
        },
        'link_two': {
            'type': 'string',
            'linkTo': ITEM_E
        },
    }


@pytest.fixture
def test_parent_type_object_schema():
    """ Intended schema structure to test schemas that have an object or array of object field
        that contains a linkTo.
    """
    return {
        'single_object': {
            'type': 'object',
            'properties': {
                'alias': {
                    'type': 'string'
                },
                'link_one': {
                    'type': 'string',
                    'linkTo': ITEM_C
                }
            }
        },
        'many_objects': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'alias': {
                        'type': 'string'
                    },
                    'link_two': {
                        'type': 'string',
                        'linkTo': ITEM_E
                    }
                }
            }
        }
    }


@pytest.mark.skip  # mocks are not setup to handle the new method
class TestInvalidationScopeUnit:

    # actual snovault types, used in basic test
    TESTING_LINK_TARGET_SNO = 'TestingLinkTargetElasticSearch'
    TESTING_LINK_TARGET_ELASTICSEARCH = 'TestingLinkTargetElasticSearch'

    @staticmethod
    def build_invalidated_and_secondary_objects(number_to_generate, item_type):
        """ Helper that returns 2-tuple of invalidated,secondary intermediary structure

        :param number_to_generate: number of uuids to generate
        :param item_type: what type they should be
        :return: invalidated (list), secondary (set)
        """
        invalidated = []
        secondary = set()
        for i in range(number_to_generate):
            _id = 'uuid' + str(i)
            invalidated.append((_id, item_type))
            secondary.add(_id)
        return invalidated, secondary

    @staticmethod
    def run_test_and_reset_secondary(registry, diff, invalidated, secondary, expected):
        """ Helper method that copies and filters the secondary list, checking that the
            expected number of uuids are there.
        """
        secondary_copy = copy.deepcopy(secondary)
        filter_invalidation_scope(registry, diff, invalidated, secondary_copy)
        assert len(secondary_copy) == expected

    @pytest.mark.parametrize('item_type', [TESTING_LINK_TARGET_SNO, TESTING_LINK_TARGET_ELASTICSEARCH])
    def test_invalidation_scope_basic(self, testapp, item_type):
        """ Uses some test data to test base invalidation scope cases """
        registry = testapp.app.registry
        diff = [
            item_type + '.name'
        ]
        invalidated = [  # not touched so can remain the same
            (UUID1, 'TestingLinkSourceSno')
        ]
        secondary = {UUID1}
        # since name field is not embedded, we don't care about this edit
        self.run_test_and_reset_secondary(registry, diff, invalidated, secondary, 0)

        # override diff to instead correspond to a modification to status field, which is embedded
        diff = [item_type + '.status']
        self.run_test_and_reset_secondary(registry, diff, invalidated, secondary, 1)

        # given 10 uuids of the same type, it should still not filter any out
        invalidated, secondary = self.build_invalidated_and_secondary_objects(10, item_type)
        filter_invalidation_scope(registry, diff, invalidated, secondary)
        assert len(secondary) == 10

    @pytest.mark.parametrize('item_type,embedded_field', [(ITEM_A, ['link_one.name']), (ITEM_B, ['link_one.key'])])
    def test_invalidation_scope_complicated(self, testapp, item_type, test_link_source_schema, embedded_field):
        """ Runs two scenarios:
                1. We edited field 'name' on ITEM_C, invalidating ITEM_A since it embeds 'name'
                2. We edited field 'name' on ITEM_C, NOT invalidating ITEM_B since it embeds 'key'
        """
        registry = testapp.app.registry
        diff = [  # this diff should only invalidate Item_A
            ITEM_C + '.name'
        ]
        invalidated = [
            (UUID1, item_type)
        ]
        secondary = {UUID1}
        with invalidation_scope_mocks(schema=test_link_source_schema, embedded_list=embedded_field):
            filter_invalidation_scope(registry, diff, invalidated, secondary)
            if embedded_field[0] == 'link_one.name':
                assert len(secondary) == 1
            else:
                assert len(secondary) == 0

    @pytest.mark.parametrize('diff,embedded_list,expected',
                             [([ITEM_C + '.value'], ['single_object.link_one.value'], 1),  # value matches
                              ([ITEM_C + '.name'], ['single_object.link_one.value'], 0),  # name does not match embed
                              ([ITEM_E + '.name'], ['many_objects.link_two.name'], 1),  # name matches
                              ([ITEM_E + '.name'], ['many_objects.link_two.*'], 1),  # * matches
                              ([ITEM_C + '.*'], ['many_objects.link_two.*'], 0),  # * does not match embed
                              ([ITEM_E + '.value'], ['many_objects.link_two.name'], 0),  # value does not match embed
                              ([ITEM_E + '.value'], ['single_object.link_one.value'], 0),  # diff does not match field
                              ([ITEM_C + '.name'], ['single_object.link_one.*'], 1),  # * matches
                              ([ITEM_C + '.name'], ['many_objects.link_two.*'], 0),  # * does not match embed
                              ([ITEM_C + '.name', ITEM_C + '.value', ITEM_C + '.classification'],
                               ['single_object.link_one.classification'], 1),  # larger diff
                              ([ITEM_C + '.name', ITEM_C + '.value', ITEM_C + '.classification'],
                               ['many_objects.link_two.classification'], 0),  # larger diff w/o embed
                              ([ITEM_E + '.name', ITEM_E + '.value', ITEM_E + '.classification'],
                               ['many_objects.link_two.classification'], 1),
                              ])
    def test_invalidation_scope_object(self, testapp, test_parent_type_object_schema, diff, embedded_list, expected):
        """ Tests that modifying an item that is an object-like linkTo remains in the invalidation scope. """
        registry = testapp.app.registry
        invalidated = [
            (UUID1, 'dummy_type')  # this type doesn't matter
        ]
        secondary = {UUID1}
        with invalidation_scope_mocks(schema=test_parent_type_object_schema, embedded_list=embedded_list):
            self.run_test_and_reset_secondary(registry, diff, invalidated, secondary, expected)

    @pytest.mark.parametrize('item_type,embedded_list',
                             [(ITEM_A, ['link_one.name', 'link_two.key', 'link_three.value']),
                              (ITEM_B, ['link_one.key', 'link_two.value', 'link_three.name'])])
    def test_invalidation_scope_many_unseen_changes(self, testapp, item_type, embedded_list, test_link_source_schema):
        """ Similar test to above except multiple modifications are made in the diff,
            NONE of which should impact the invalidation scope.
        """
        registry = testapp.app.registry
        diff = [  # this diff should invalidate NONE, so secondary should be cleared
            ITEM_C + '.value',
            ITEM_D + '.name',
            ITEM_E + '.key'
        ]
        invalidated = [
            (UUID1, item_type)
        ]
        secondary = {UUID1}
        with invalidation_scope_mocks(schema=test_link_source_schema, embedded_list=embedded_list):
            self.run_test_and_reset_secondary(registry, diff, invalidated, secondary, 0)

    @pytest.mark.parametrize('item_type,embedded_list', [
        (ITEM_A, ['link_one.name', 'link_two.key', 'link_three.value']),
        (ITEM_B, ['link_one.key', 'link_two.value', 'link_three.name'])])
    def test_invalidation_scope_many_changes(self, testapp, item_type, embedded_list, test_link_source_schema):
        """ Again similar to the above test except this time the diff will trigger an
            invalidation on both item types.
        """
        registry = testapp.app.registry
        diff = [  # this diff invalidates BOTH
            ITEM_C + '.name',
            ITEM_D + '.value'
        ]
        invalidated = [
            (UUID1, item_type)
        ]
        secondary = {UUID1}
        with invalidation_scope_mocks(schema=test_link_source_schema, embedded_list=embedded_list):
            self.run_test_and_reset_secondary(registry, diff, invalidated, secondary, 1)

    @pytest.mark.parametrize('diff,embedded_list',
                             [([ITEM_C + '.name', 'Item_X.value'], ['link_one.value']),
                              ([ITEM_D + '.value', ITEM_E + '.key'], ['link_two.name', 'link_three.value']),
                              ([ITEM_C + '.key', ITEM_D + '.key', ITEM_E + '.key'], ['link_one.value', 'link_two.name'])])
    def test_invalidation_scope_negative_diffs(self, testapp, test_link_source_schema, diff, embedded_list):
        """ Tests a few possible edit + embedding_list combinations that should NOT result
            in re-indexing (so the added item_type is removed from the invalidation scope).
        """
        registry = testapp.app.registry
        invalidated = [
            (UUID1, ITEM_A)  # item type doesn't matter here
        ]
        secondary = {UUID1}
        with invalidation_scope_mocks(schema=test_link_source_schema, embedded_list=embedded_list):
            self.run_test_and_reset_secondary(registry, diff, invalidated, secondary, 0)

    @pytest.mark.parametrize('diff,embedded_list',
                             [([ITEM_C + '.value'], ['link_one.value']),
                              ([ITEM_C + '.value'], ['link_one.*']),
                              ([ITEM_D + '.key', ITEM_F + '.value'], ['link_two.key', 'link_four.key'])])
    def test_invalidation_scope_positive_diffs(self, testapp, test_link_source_schema, diff, embedded_list):
        """ Tests a few possible edit + embedding_list combinations that should result in
            invalidation.
        """
        registry = testapp.app.registry
        invalidated = [
            (UUID1, ITEM_A)  # item type doesn't matter here
        ]
        secondary = {UUID1}
        with invalidation_scope_mocks(schema=test_link_source_schema, embedded_list=embedded_list):
            self.run_test_and_reset_secondary(registry, diff, invalidated, secondary, 1)

    @pytest.mark.parametrize('diff,embedded_list,base_types,child_types', [
        ([ITEM_D + '.value'], ['link_one.value'], [ITEM_C], [ITEM_D])
    ])
    def test_invalidation_scope_base_types(self, testapp, diff, embedded_list, base_types, child_types,
                                           test_parent_type_schema):
        """ Runs some tests that involve base type resolution """
        registry = testapp.app.registry
        invalidated = [
            (UUID1, ITEM_A)
        ]
        secondary = {UUID1}
        with invalidation_scope_mocks(schema=test_parent_type_schema, embedded_list=embedded_list,
                                      base_types=base_types, child_types=child_types):
            self.run_test_and_reset_secondary(registry, diff, invalidated, secondary, 1)

    def test_invalidation_scope_get_child_types(self, testapp):
        """ Tests that we can resolve child types correctly. """
        item_child_types = determine_child_types(testapp.app.registry, 'Item')
        assert item_child_types == ['AbstractItemTestSecondSubItem', 'AbstractItemTestSubItem',  # all types defined
                                    'EmbeddingTest', 'NestedEmbeddingContainer', 'NestedObjectLinkTarget',
                                    'TestingBiogroupSno', 'TestingBiosampleSno', 'TestingBiosourceSno',
                                    'TestingCalculatedProperties', 'TestingDependencies', 'TestingDownload',
                                    'TestingIndividualSno', 'TestingLinkAggregateSno', 'TestingLinkSourceSno',
                                    'TestingLinkTargetElasticSearch', 'TestingLinkTargetSno', 'TestingMixins',
                                    'TestingNestedEnabled', 'TestingPostPutPatchSno', 'TestingServerDefault']
        abstract_item_type_child_types = determine_child_types(testapp.app.registry, 'AbstractItemTest')
        assert abstract_item_type_child_types == ['AbstractItemTestSecondSubItem', 'AbstractItemTestSubItem']


###########################################################
# Tests based on actual views defined in testing_views.py #
###########################################################


@pytest.fixture
def invalidation_scope_individual_data():
    return [
        {
            'full_name': 'Jane Doe',
            'uid': 'abc123',
            'specimen': 'blood'
        },
        {
            'full_name': 'Unknown Contributor',
            'uid': 'UNK1',
            'specimen': 'skin'
        },
    ]


@pytest.fixture
def invalidation_scope_biosample_data():
    return [
        {
            'identifier': 'SNOID123',
            'quality': 100,
            'ranking': 1,
            'alias': '123',
            'contributor': 'Unknown Contributor',
            'technical_reviews': [
                'SNONOTE1', 'SNONOTE2'
            ]
        },
        {
            'identifier': 'SNOID456',
            'quality': 98,
            'ranking': 2,
            'alias': '456',
            'contributor': 'Jane Doe'
        },
        {
            'identifier': 'SNOID789',
            'quality': 95,
            'ranking': 3,
            'alias': '789',
            'contributor': 'Jane Doe'
        }
    ]


@pytest.fixture
def invalidation_scope_biosource_data():
    return [
        {
            'identifier': 'SNOIDABC',
            'samples': ['SNOID123', 'SNOID456', 'SNOID789']
        },
        {
            'identifier': 'SNOIDDEF',
            'samples': ['SNOID456', 'SNOID789'],
            'sample_objects': [
                {
                    'notes': 'A note about SNOID456',
                    'associated_sample': 'SNOID456'
                }
            ],
            'contributor': 'Jane Doe'
        }
    ]


@pytest.fixture
def invalidation_scope_biogroup_data():
    return [{
        'name': 'test-group',
        'sources': 'SNOIDDEF'
    }]


@pytest.fixture
def invalidation_scope_note_data():
    return [
        {
            'assessment': {
                'call': True,
                'classification': 'Present'
            },
            'identifier': 'SNONOTE1'
        },
        {
            'assessment': {
                'call': False,
                'classification': 'Repeat Region'
            },
            'identifier': 'SNONOTE2',
            'superseding_note': 'SNONOTE1'
        },
    ]


@pytest.fixture
def invalidation_scope_workbook(testapp, invalidation_scope_biosample_data, invalidation_scope_biosource_data,
                                invalidation_scope_biogroup_data, invalidation_scope_individual_data,
                                invalidation_scope_note_data):
    """ Posts 2 individuals, 3 biosamples, 2 biosources and 1 biogroup for integrated testing. """
    groups, sources, samples, notes = [], [], [], []
    for note in invalidation_scope_note_data:
        res = testapp.post_json('/TestingNoteSno', note, status=201).json['@graph'][0]
        notes.append(res)
    for indiv in invalidation_scope_individual_data:
        testapp.post_json('/TestingIndividualSno', indiv, status=201)
    for biosample in invalidation_scope_biosample_data:
        res = testapp.post_json('/TestingBiosampleSno', biosample, status=201).json['@graph'][0]
        samples.append(res)
    for biosource in invalidation_scope_biosource_data:
        res = testapp.post_json('/TestingBiosourceSno', biosource, status=201).json['@graph'][0]
        sources.append(res)
    for biogroup in invalidation_scope_biogroup_data:
        res = testapp.post_json('/TestingBiogroupSno', biogroup, status=201).json['@graph'][0]
        groups.append(res)
    return groups, sources, samples, notes


class TestingInvalidationScopeIntegrated:

    @staticmethod
    def runtest(testapp, diff, invalidated, secondary, expected):
        __test__ = False
        filter_invalidation_scope(testapp.app.registry, diff, invalidated, secondary)
        assert len(secondary) == expected

    def test_invalidation_scope_integrated_simple_modification(self, testapp, invalidation_scope_workbook):
        """ Integrated test that simulated patching the identifier field on biosample. Because identifier
            is an embed for biosource (direct) and also an embed for biogroup (*), all 3 items should be
            invalidated.
        """
        groups, sources, samples, _ = invalidation_scope_workbook
        diff = ['TestingBiosampleSno.identifier']
        invalidated = [(obj['@id'], obj['@type'][0]) for obj in sources + groups]
        secondary = {obj['@id'] for obj in sources + groups}
        self.runtest(testapp, diff, invalidated, secondary, 3)

    def test_invalidation_scope_integrated_many_modifications(self, testapp, invalidation_scope_workbook):
        """ Integrated test that simulated patching the identifier field on biosample. Because identifier
            is an embed for biosource (direct) and also an embed for biogroup (*), all 3 items should be
            invalidated.
        """
        groups, sources, samples, _ = invalidation_scope_workbook
        diff = ['TestingBiosampleSno.alias', 'TestingBiosampleSno.ranking', 'TestingBiosampleSno.quality']
        invalidated = [(obj['@id'], obj['@type'][0]) for obj in sources + groups]
        secondary = {obj['@id'] for obj in sources + groups}
        self.runtest(testapp, diff, invalidated, secondary, 3)  # quality + * will pick up all

    def test_invalidation_scope_integrated_partly_invisible_modification(self, testapp, invalidation_scope_workbook):
        """ Integrated test that simulates patching the ranking field on biosample. This field is not a
            direct embed on biosource, so those items should not be invalidated - only biogroup.
        """
        groups, sources, samples, _ = invalidation_scope_workbook
        diff = ['TestingBiosampleSno.ranking']
        invalidated = [(obj['@id'], obj['@type'][0]) for obj in sources + groups]
        secondary = {obj['@id'] for obj in sources + groups}
        self.runtest(testapp, diff, invalidated, secondary, 1)

    def test_invalidation_scope_integrated_wholly_invisible_modification(self, testapp, invalidation_scope_workbook):
        """ Integrated test that simulates patching the identifier field on biosource. This field is not a
            direct embed on biogroup, so nothing is invalidated.
        """
        groups, sources, samples, _ = invalidation_scope_workbook
        diff = ['TestingBiosourceSno.identifier']
        invalidated = [(obj['@id'], obj['@type'][0]) for obj in groups]
        secondary = {obj['@id'] for obj in groups}
        self.runtest(testapp, diff, invalidated, secondary, 1)  # answer is 1 because of default_diff

    def test_invalidation_scope_integrated_depth3_modification_matches(self, testapp, invalidation_scope_workbook):
        """ Integrated test that simulates patches the alias on biosource. This field is a direct embed on biosource
            under 'sample_objects.associated_sample.alias' - thus only biosources should be invalidated - but since
            we cannot differentiate this edit by field, all 3 are invalidated.
        """
        groups, sources, samples, _ = invalidation_scope_workbook
        diff = ['TestingBiosampleSno.alias']
        invalidated = [(obj['@id'], obj['@type'][0]) for obj in sources + groups]
        secondary = {obj['@id'] for obj in sources + groups}
        self.runtest(testapp, diff, invalidated, secondary, 3)

    def test_invalidation_scope_integrated_depth4_modification_partial(self, testapp, invalidation_scope_workbook):
        """ Integrated test that simulates patches the specimen on individual. Since only biosample embeds this
            via *, so biosources should be pruned.
        """
        groups, sources, samples, _ = invalidation_scope_workbook
        diff = ['TestingIndividualSno.specimen']
        invalidated = [(obj['@id'], obj['@type'][0]) for obj in sources + samples]
        secondary = {obj['@id'] for obj in sources + samples}
        self.runtest(testapp, diff, invalidated, secondary, 3)

    def test_invalidation_scope_integrated_depth4_modification_full(self, testapp, invalidation_scope_workbook):
        """ Integrated test that simulates patches the full_name on individual. """
        groups, sources, samples, _ = invalidation_scope_workbook
        diff = ['TestingIndividualSno.full_name']
        invalidated = [(obj['@id'], obj['@type'][0]) for obj in sources + samples]
        secondary = {obj['@id'] for obj in sources + samples}
        self.runtest(testapp, diff, invalidated, secondary, 2)

    @pytest.mark.parametrize('diff,expected',
                             [(['TestingIndividualSno.uid'], 1),  # only biogroup is in scope
                              (['TestingIndividualSno.full_name'], 3),  # biosource + biogroup
                              (['TestingIndividualSno.specimen'], 4),  # biosamples + biogroup
                              (['TestingIndividualSno.uid', 'TestingIndividualSno.specimen',
                                'TestingIndividualSno.full_name'], 6),  # all care
                              ])
    def test_invalidation_scope_integrated_all(self, testapp, invalidation_scope_workbook, diff, expected):
        """ Integrated test that simulates many diffs checking that the right # of items were invalidated. """
        groups, sources, samples, _ = invalidation_scope_workbook
        invalidated = [(obj['@id'], obj['@type'][0]) for obj in groups + sources + samples]
        secondary = {obj['@id'] for obj in groups + sources + samples}
        self.runtest(testapp, diff, invalidated, secondary, expected)

    def test_invalidation_scope_integrated_link_subobject(self, testapp, invalidation_scope_workbook):
        """ Tests patching biosource.sample_object.notes, which should invalidate """
        groups, _, _, _ = invalidation_scope_workbook
        diff = ['TestingBiosourceSno.sample_objects.notes']
        invalidated = [(obj['@id'], obj['@type'][0]) for obj in groups]
        secondary = {obj['@id'] for obj in groups}
        with type_embedded_list_mock(embedded_list=['sources.sample_objects.notes']):
            self.runtest(testapp, diff, invalidated, secondary, 1)

    @pytest.mark.parametrize('diff,expected', [
        (['TestingNoteSno.superseding_note'], 2),
        (['TestingNoteSno.assessment'], 5),
        (['TestingNoteSno.assessment.call'], 5),
        (['TestingNoteSno.assessment.date_call_made'], 0),
        (['TestingNoteSno.review'], 3),
        (['TestingNoteSno.review.date_reviewed'], 3),
        (['TestingNoteSno.previous_note'], 0),
    ])
    def test_invalidation_scope_integrated_multiple_links(self, testapp, invalidation_scope_workbook, diff, expected):
        groups, sources, samples, notes = invalidation_scope_workbook
        invalidated = [(obj['@id'], obj['@type'][0]) for obj in groups + sources + samples + notes]
        secondary = {obj['@id'] for obj in groups + sources + samples + notes}
        self.runtest(testapp, diff, invalidated, secondary, expected)

    def test_invalidation_scope_default_diff(self, testapp, invalidation_scope_workbook):
        """ Integrated test verifies that default_diff is added correctly
            ie: diff should not trigger invalidation, but default_diff will
        """
        groups, sources, samples, _ = invalidation_scope_workbook
        invalidated = [(obj['@id'], obj['@type'][0]) for obj in groups + sources + samples]
        secondary = {obj['@id'] for obj in groups + sources + samples}
        self.runtest(testapp, ['TestingBiosourceSno.identifier'], invalidated, secondary, 1)

    def test_invalidation_scope_empty_patch(self, testapp, invalidation_scope_workbook):
        """ Integrated test verifies that default_diff is added correctly
            ie: diff should not trigger invalidation, but default_diff will
        """
        groups, sources, samples, _ = invalidation_scope_workbook
        invalidated = [(obj['@id'], obj['@type'][0]) for obj in groups + sources + samples]
        secondary = {obj['@id'] for obj in groups + sources + samples}
        self.runtest(testapp, [], invalidated, secondary, 0)



import pytest

from copy import deepcopy
from ..elasticsearch.cached_views import filter_embedded


def test_filtered_embedded_admin(embedded_lab, effective_princ_admin):
    original = deepcopy(embedded_lab)
    filtered = filter_embedded(embedded_lab, effective_princ_admin)
    assert filtered == embedded_lab
    assert original == embedded_lab


def test_filtered_embedded_no_perms(embedded_lab):
    original = deepcopy(embedded_lab)
    filter_embedded(embedded_lab, [])
    assert original != embedded_lab
    assert embedded_lab == {'lab': {'error': "no view permissions"}}


def test_filtered_embedded_second_level_no_perms(embedded_2nd, effective_princ_not_logged_in):
    original = deepcopy(embedded_2nd)
    filter_embedded(embedded_2nd, effective_princ_not_logged_in)
    assert original != embedded_2nd
    assert embedded_2nd['lab']['award'] == {'error': "no view permissions"}


def test_filtered_embedded_array(embedded_3rd_array, effective_princ_lab):
    modified = deepcopy(embedded_3rd_array)
    original = deepcopy(modified)

    filter_embedded(modified, effective_princ_lab)
    assert modified != original
    # no perms for this one
    assert modified['biosample']['biosource'][1] == {'error': "no view permissions"}
    # but we can see this one
    assert 'uuid' in modified['biosample']['biosource'][0].keys()
    assert 'uuid' in modified['biosample']['biosource'][0]['individual']['organism'].keys()


@pytest.fixture()
def effective_princ_not_logged_in():
    return ['system.Everyone']


@pytest.fixture()
def effective_princ_lab():
    return ['lab.labtest', 'system.Everyone']


@pytest.fixture()
def embedded_2nd():
    return {'lab': {'fax': '000-000-0000', 'city': 'Boston', 'name': 'test-4dn-lab',
                    'state': 'MA', '@type': ['Lab', 'Item'],
                    'uuid': '828cd4fe-ebb0-4b22-a94a-d2e3a36cc988',
                    'display_title': '4DN Testing Lab',
                    'principals_allowed': {'view': ['system.Everyone']},
                    'award': {'name': 'test_award',
                              'principals_allowed': {'view': ['group.admin']}
                              }
                    }
            }


@pytest.fixture()
def embedded_3rd_array():
    return {'biosample': {
        'status': 'in review by lab',
        'accession': '4DNBS2548734',
        'biosource': [
            {
                'status': 'in review by lab',
                'accession': '4DNSR9R56R7W',
                'individual': {
                    'organism': {
                        '@id': '/organisms/human/',
                        '@type': [
                            'Organism',
                            'Item'
                        ],
                        'uuid': '7745b647-ff15-4ff3-9ced-b897d4e2983c',
                        'principals_allowed': {
                            'view': [
                                'system.Everyone'
                            ],
                        }
                    }
                },
                'schema_version': '2',
                '@id': '/biosources/4DNSR9R56R7W/',
                '@type': [
                    'Biosource',
                    'Item'
                ],
                'uuid': '331111bc-8535-4448-903e-854af460b254',
                'principals_allowed': {
                    'view': [
                        'group.admin',
                        'lab.labtest',
                    ],
                },
'biosource_name': 'GM12878'
},

            {
                'uuid': '331111bc-8535-4448-903e-854af460b254',
                'principals_allowed': {
                    'view': [
                        'group.admin',
                    ],
                    },
                'biosource_name': 'dumb, biosource'
}
],
'treatments': [
    {
        'url': 'https://www.cyclingnews.com',
        'status': 'in review by lab',
        'aliases': [
            'Awesome:Treatment'
        ],
        'rnai_type': 'siRNA',
        'description': 'RNAi treatment for rS3',
        'date_created': '2017-07-31T14:31:24.815183+00:00',
        'schema_version': '1',
        'target_sequence': 'ATGCATGC',
        '@id': '/treatments-rnai/686b362f-4eb6-4a9c-8173-3ab267307e3b/',
        '@type': [
            'TreatmentRnai',
            'Treatment',
            'Item'
        ],
        'uuid': '686b362f-4eb6-4a9c-8173-3ab267307e3b',
        'display_title': 'TreatmentRnai from 2017-07-31',
        'principals_allowed': {
            'view': [
                'award.b0b9c607-f8b4-4f02-93f4-9895b461334b',
                'group.admin',
                'group.read-only-admin',
                'lab.828cd4fe-ebb0-4b36-a94a-d2e3a36cc989',
                'remoteuser.EMBED',
                'remoteuser.INDEXER'
            ],
            'edit': [
                'group.admin',
                'submits_for.828cd4fe-ebb0-4b36-a94a-d2e3a36cc989'
            ]
        },
        'treatment_type': 'siRNA for Gene:PARK2, FMR1'
    },
    {
        'url': 'https://www.cyclingnews.com',
        'status': 'in review by lab',
        'aliases': [
            'Awesome:Treatment2'
        ],
        'rnai_type': 'shRNA',
        'description': 'RNAi treatment for rS5',
        'date_created': '2017-07-31T14:31:24.873541+00:00',
        'schema_version': '1',
        'target_sequence': 'ATGCATGC',
        '@id': '/treatments-rnai/686b362f-4eb6-4a9c-8173-3ab267307a8d/',
        '@type': [
            'TreatmentRnai',
            'Treatment',
            'Item'
        ],
        'uuid': '686b362f-4eb6-4a9c-8173-3ab267307a8d',
        'display_title': 'TreatmentRnai from 2017-07-31',
        'principals_allowed': {
            'view': [
                'award.b0b9c607-f8b4-4f02-93f4-9895b461334b',
                'group.admin',
                'group.read-only-admin',
                'lab.828cd4fe-ebb0-4b36-a94a-d2e3a36cc989',
                'remoteuser.EMBED',
                'remoteuser.INDEXER'
            ],
            'edit': [
                'group.admin',
                'submits_for.828cd4fe-ebb0-4b36-a94a-d2e3a36cc989'
            ]
        },
        'treatment_type': 'shRNA for Gene:PARK2, FMR1'
    }
    ],
        'description': 'GM12878 prepared for Hi-C, batch 2',
        'date_created': '2017-07-31T14:31:25.014388+00:00',
        'modifications': [
    {
        'url': 'https://www.cyclingnews.com',
        'status': 'in review by lab',
        'description': 'Stable Tranfection of GFP construct',
        'date_created': '2017-07-31T14:31:23.390224+00:00',
        'schema_version': '1',
        'modification_type': 'Stable Transfection',
        '@id': '/modifications/431106bc-8535-4448-903e-854af460b265/',
        '@type': [
            'Modification',
            'Item'
        ],
        'uuid': '431106bc-8535-4448-903e-854af460b265',
        'display_title': 'Stable Transfection for Gene:PARK2, FMR1',
        'principals_allowed': {
            'view': [
                'award.b0b9c607-f8b4-4f02-93f4-9895b461334b',
                'group.admin',
                'group.read-only-admin',
                'lab.828cd4fe-ebb0-4b36-a94a-d2e3a36cc989',
                'remoteuser.EMBED',
                'remoteuser.INDEXER'
            ],
            'edit': [
                'group.admin',
                'submits_for.828cd4fe-ebb0-4b36-a94a-d2e3a36cc989'
            ]
        },
        'modification_name': 'Stable Transfection for Gene:PARK2, FMR1',
        'modification_name_short': 'Stable Transfection for Gene:PARK2, FMR1'
    },
    {
        'url': 'https://www.cyclingnews.com',
        'status': 'in review by lab',
        'description': 'Other type of modification',
        'date_created': '2017-07-31T14:31:23.419011+00:00',
        'schema_version': '1',
        'modification_type': 'Other',
        '@id': '/modifications/431106bc-8535-4448-903e-854af460b276/',
        '@type': [
            'Modification',
            'Item'
        ],
        'uuid': '431106bc-8535-4448-903e-854af460b276',
        'display_title': 'Other',
        'principals_allowed': {
            'view': [
                'group.admin',
                'group.read-only-admin',
                'lab.828cd4fe-ebb0-4b36-a94a-d2e3a36cc989',
                'remoteuser.EMBED',
                'remoteuser.INDEXER'
            ],
            'edit': [
                'group.admin',
                'submits_for.828cd4fe-ebb0-4b36-a94a-d2e3a36cc989'
            ]
        },
        'modification_name': 'Other',
        'modification_name_short': 'Other'
    }
],
'schema_version': '1',
'@id': '/biosamples/4DNBS2548734/',
'@type': [
    'Biosample',
    'Item'
],
'uuid': '231111bc-8535-4448-903e-854af460b242',
'display_title': '4DNBS2548734',
'principals_allowed': {
    'view': [
        'group.admin',
        'lab.labtest',
    ],
},
'biosource_summary': 'GM12878'
}
}


@pytest.fixture()
def embedded_lab():
    return {'lab': {'fax': '000-000-0000', 'city': 'Boston', 'name': 'test-4dn-lab', 'state': 'MA', 'title': '4DN Testing Lab', 'status': 'current', 'country': 'USA', 'address1': 'Biomedical Bioinfomatics', 'address2': '10 Schattuck Street', 'postal_code': '02115', 'date_created': '2017-07-31T14:31:13.532200+00:00', 'institute_name': 'Harvard Medical School', 'schema_version': '1', 'institute_label': 'HMS', '@id': '/labs/test-4dn-lab/', '@type': ['Lab', 'Item'], 'uuid': '828cd4fe-ebb0-4b22-a94a-d2e3a36cc988', 'display_title': '4DN Testing Lab', 'principals_allowed': {'view': ['system.Everyone'], 'edit': ['group.admin', 'submits_for.828cd4fe-ebb0-4b22-a94a-d2e3a36cc988']}}}


@pytest.fixture()
def effective_princ_admin():
    return ['auth0.4dndcic@gmail.com', 'system.Everyone', 'system.Authenticated', 'submits_for.828cd4fe-ebb0-4b36-a94a-d2e3a36cc989', 'lab.828cd4fe-ebb0-4b36-a94a-d2e3a36cc989', 'userid.986b362f-4eb6-4a9c-8173-3ab267307e3a', 'group.submitter', 'award.b0b9c607-f8b4-4f02-93f4-9895b461334b', 'group.admin', 'remoteuser.EMBED', 'viewing_group.4DN']



import os
import pytest
import mock

from ..interfaces import TYPES
from ..elasticsearch.create_mapping import (
    merge_schemas,
    type_mapping,
    update_mapping_by_embed,
    get_items_to_upgrade,
)
from ..elasticsearch.interfaces import ELASTIC_SEARCH
from .test_views import PARAMETERIZED_NAMES
from ..settings import Settings
from ..util import add_default_embeds
from contextlib import contextmanager


unit_test_type = 'EmbeddingTest'


@contextmanager
def mappings_use_nested(value=True):
    """ Context manager that sets the MAPPINGS_USE_NESTED setting with the given value, default True """
    old_setting = Settings.MAPPINGS_USE_NESTED
    try:
        Settings.MAPPINGS_USE_NESTED = value
        yield
    finally:
        Settings.MAPPINGS_USE_NESTED = old_setting


@pytest.mark.parametrize('item_type', PARAMETERIZED_NAMES)
def test_type_mapping(registry, item_type):
    """
    Test basic mapping properties for each item type
    """
    with mappings_use_nested(False):
        mapping = type_mapping(registry[TYPES], item_type)
        assert mapping
        assert 'properties' in mapping
        if item_type == 'TestingLinkTargetElasticSearch':
            assert mapping['properties']['reverse_es'].get('type', 'object') != 'nested'  # should not occur here

        # check calculated properties on objects/arrays of objects are mapped correctly
        if item_type == 'TestingCalculatedProperties':
            assert mapping['properties']['nested']['properties']['key']['type'] == 'text'
            assert mapping['properties']['nested']['properties']['value']['type'] == 'text'
            assert mapping['properties']['nested']['properties']['keyvalue']['type'] == 'text'
            assert mapping['properties']['nested2']['properties']['key']['type'] == 'text'
            assert mapping['properties']['nested2']['properties']['value']['type'] == 'text'
            assert mapping['properties']['nested2']['properties']['keyvalue']['type'] == 'text'


def test_type_mapping_nested(registry):
    """
    Tests that mapping a field with a list of dicts in it maps with type=nested only if told to do so on
    the schema. For this case it is not specified, so if object is expected.
    """
    with mappings_use_nested(True):
        mapping = type_mapping(registry[TYPES], 'TestingLinkTargetElasticSearch')
        assert mapping
        assert 'properties' in mapping
        # if type is defined on this field, it should beg object, NOT nested since it is not enabled on this field
        assert mapping['properties']['reverse_es'].get('type', 'object') == 'object'


def test_type_mapping_nested_with_disabled_parameter(registry):
    """ Tests that mapping a type with an object field with nested enabled correctly maps
        with nested.
    """
    with mappings_use_nested(True):
        mapping = type_mapping(registry[TYPES], 'TestingNestedEnabled')
        assert mapping
        assert 'properties' in mapping
        assert mapping['properties']['object_options'].get('type', 'object') != 'nested'  # neither enabled
        assert mapping['properties']['disabled_array_of_objects_in_calc_prop'].get('type', 'object') != 'nested'
        assert mapping['properties']['enabled_array_of_objects_in_calc_prop']['type'] == 'nested'  # enabled
        assert mapping['properties']['nested_options']['type'] == 'nested'  # enabled


def test_merge_schemas(registry):
    """ Tests merging schemas with EmbeddingTest """
    test_schema = registry[TYPES][unit_test_type].schema
    test_subschema = test_schema['properties']['attachment']
    res = merge_schemas(test_subschema, registry[TYPES])
    assert res
    assert res != test_subschema
    assert res['properties']['attachment']['attachment'] is True


def test_update_mapping_by_embed(registry):
    # first, test with dummy data
    curr_s = {'title': 'Test', 'type': 'string'}
    curr_e = 'test'
    curr_m = {'properties': {}}
    new_m = update_mapping_by_embed(curr_m, curr_e, curr_s)
    assert 'test' in curr_m['properties']
    assert new_m['type'] == 'text'
    assert 'raw' in new_m['fields']
    assert 'lower_case_sort' in new_m['fields']

    # then test with real data and wildcard (*)
    test_schema = registry[TYPES][unit_test_type].schema
    test_subschema = test_schema['properties']['attachment']
    curr_s = merge_schemas(test_subschema, registry[TYPES])
    # * means embed all top level fields
    curr_e = '*'
    curr_m = {'properties': {}}
    new_m = update_mapping_by_embed(curr_m, curr_e, curr_s)
    for s_key in curr_s['properties']:
        check = curr_s['properties'][s_key]
        # this syntax needed for linkTos which could be arrays
        if 'linkTo' not in check.get('items', check):
            assert s_key in new_m['properties']


# types to test
TEST_TYPES = ['testing_mixins', 'embedding_test', 'nested_embedding_container', 'nested_object_link_target',
         'testing_download', 'testing_link_source_sno', 'testing_link_aggregate_sno', 'testing_link_target_sno',
         'testing_post_put_patch_sno', 'testing_dependencies', 'testing_link_target_elastic_search']


@pytest.mark.parametrize('item_type', TEST_TYPES)
def test_create_mapping_correctly_maps_embeds(registry, item_type):
    """
    This test does not actually use elasticsearch
    Only tests the mappings generated from schemas
    This test existed in FF/CGAP and has been ported here so we can detect issues earlier
    """
    mapping = type_mapping(registry[TYPES], item_type)
    assert mapping
    type_info = registry[TYPES].by_item_type[item_type]
    schema = type_info.schema
    embeds = add_default_embeds(item_type, registry[TYPES], type_info.embedded_list, schema)
    # assert that all embeds exist in mapping for the given type
    for embed in embeds:
        mapping_pointer = mapping
        split_embed = embed.split('.')
        for idx, split_ in enumerate(split_embed):
            # see if this is last level of embedding- may be a field or object
            if idx == len(split_embed) - 1:
                if 'properties' in mapping_pointer and split_ in mapping_pointer['properties']:
                    final_mapping = mapping_pointer['properties']
                else:
                    final_mapping = mapping_pointer
                if split_ != '*':
                    assert split_ in final_mapping
                else:
                    assert 'properties' in final_mapping or final_mapping.get('type') == 'object'
            else:
                assert split_ in mapping_pointer['properties']
                mapping_pointer = mapping_pointer['properties'][split_]


def test_create_mapping_drops_unmappable_properties(registry):
    """Use EmbeddingTest schema to ensure unmappable properties under
    patternProperties and additionalProperties are not mapped.
    """
    test_item_type = "embedding_test"
    expected_embeds = ["pattern_property_embed", "additional_property_embed"]
    expected_mapped_property = "should_be_mapped"
    mapping = type_mapping(registry[TYPES], test_item_type)
    mapped_properties = mapping.get("properties")
    assert mapped_properties
    for expected_embed in expected_embeds:
        mapped_embed = mapped_properties.get(expected_embed)
        assert mapped_embed
        mapped_embed_properties = mapped_embed.get("properties", {})
        assert len(mapped_embed_properties.keys()) == 1
        assert mapped_embed_properties.get(expected_mapped_property)


@pytest.fixture
def biosample(testapp):
    url = "/testing-biosample-sno"
    item = {
        "identifier": "test_biosample",
        "ranking": 5,
    }
    return testapp.post_json(url, item, status=201).json["@graph"][0]


def mock_scan(uuids=None):
    """Mock for ES scan which returns list of dicts."""
    embedded_item = []
    if uuids:
        for uuid in uuids:
            embedded_item.append({"_id": uuid})
    mocked_scan = mock.Mock()
    mocked_scan.return_value = embedded_item 
    return mocked_scan


@mock.patch("snovault.elasticsearch.create_mapping.check_if_index_exists", return_value=True)
def test_get_items_to_upgrade(mock_check_index, testapp, biosample):
    """
    Test Elasticsearch items requiring an upgrade are identified for
    indexing.
    """
    app = testapp.app
    es = None  # ES component mocked
    item_type = "testing_biosample_sno"
    biosample_uuid = biosample["uuid"]
    with mock.patch("snovault.elasticsearch.create_mapping.scan", new=mock_scan()):
        # Mocked scan found no uuids to upgrade
        to_upgrade = get_items_to_upgrade(app, es, item_type)
        assert not to_upgrade
    with mock.patch(
        "snovault.elasticsearch.create_mapping.scan",
        new=mock_scan(uuids=[biosample_uuid])
    ):
        # Mocked scan found the biosample uuid to upgrade
        to_upgrade = get_items_to_upgrade(app, es, item_type)
        assert to_upgrade == set([biosample_uuid])
